// UTF-8 ☀☁☂☃
/*
	・アプリケーションの主要機能を持つ
	・バッファ一覧のツリーのルート
*/

package bluntirc;

import irc.*;
import base.*;
import gui.*;
import bluntirc.djava.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;

public class App
extends ConnTreeNode
{
	public static java.util.logging.Logger root_logger = java.util.logging.Logger.getLogger("");
	static{
		try{
			root_logger.setLevel(java.util.logging.Level.FINER);

			java.util.logging.FileHandler fh = new java.util.logging.FileHandler("BluntIRC-debug.log",false);
			fh.setEncoding("UTF-8");
			fh.setFormatter(new java.util.logging.SimpleFormatter());
			root_logger.addHandler(fh);
			root_logger.setFilter(null);

			java.util.logging.Handler[] hs=root_logger.getHandlers();
			for(int i=0;i<hs.length;++i)
				hs[i].setLevel(java.util.logging.Level.FINEST);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	public static java.util.logging.Logger logger = java.util.logging.Logger.getLogger("bluntirc.app");
	static{ logger.setLevel(null); }

	public static String config_AppName ="BluntIRC";
	public static String IniFileName="hash.ini";
	public String GetName(){return config_AppName;}

	// 主要な部品
	public static ImageIcon icon_app;
	public static base.Logger connection_logger= new base.Logger(){ public void Log(String s){ App.Log(s);} };
	public static OSDependence os_dependence = new OSDependence();
	public static base.SelectorRun selector;
	public static ConnectionListener conn_hook = new ConnectionListener();
	public static PrintFormatMacroManager pfm_manager = new PrintFormatMacroManager();
	public static StyleManager style_manager = new StyleManager();
	public static MenuConf menu_conf = new MenuConf();
	public static ActionManager action_manager = new ActionManager();
	public static HookManager hook_manager = new HookManager();
	public static ScriptManagerCompiled compiled_scripts = new ScriptManagerCompiled();
	public static BufferList buf_list = new BufferList();
	public static buffer.LogDocument console_doc;
	public static SelectThread select_thread;

	// ウィンドウ
	public static MainWindow main_window;
	public static ScriptManagerWindow script_manager;
	public static AllLogWindow log_all;
//	public static JobManager job_view;


	// 外部からアプリケーションのオブジェクトを取得できるようにする
	private static App application;
	public  static App getApp(){return application;}
	public static PropertyManager root_property;

	//////////////////////////////////////////////////////////////
	// 作成と破棄

	public static void main(String[] args) {
		// バージョンとカレントディレクトリを表示する
		logger.info(App.config_AppName+" build "+bdate.BuildDate.getVersion());
		logger.info("CWD is "+ (new java.io.File("nonexistent")).getAbsoluteFile().getParent() );
		// 各モジュールの初期化
		bluntirc.djava.ScriptManagerWindow.loadDummyScript();
		base.Util.AppClassLoader = App.class.getClassLoader();
		base.ConvertChar.Setup();
		irc.IRCMessageReader.init();

		// セレクタの初期化
		try{
			App.selector = new base.SelectorRun();
		}catch(java.io.IOException e){
			logger.log(java.util.logging.Level.SEVERE,"create base.SelectorRun",e);
			return;
		}

		javax.swing.SwingUtilities.invokeLater(
			new Runnable(){public void run(){
				icon_app = Util.GetImageIconFromJAR("images/app.gif");
				// 表示設定ファイルを読む
				if(! App.style_manager.load("Style.conf" ) ) return;
				if(! App.pfm_manager.load("PrintFormat.conf" ) ) return;
				console_doc=new buffer.LogDocument(App.style_manager);
				new App().init(); 
			}}
		);
	}

	public App(){ super(console_doc); }

	// 初期化と終了の間AWTスレッドを占有するわけにもいかないので、
	// Timerにステートマシンを仕込んで細切れに実行する
	private static javax.swing.Timer timer;
	private static volatile int timerState =0;
	public void init(){
		try{
			application=this; 
			setDefaultProperty();
			root_property = property;

			// BluntIRC.iniを読んでレイアウト設定だけを見る
			PropertyManager.preloadIniFile(getApp(),IniFileName);

			// アクションの登録
			action.BaseActions.entry();
			// メニューの読み込み
			menu_conf.load();

			// ウィンドウの作成
			// 全てのメッセージ
			log_all =new  AllLogWindow( !App.root_property.getBoolean("AlloLogInMain",false) );
			log_all.setSize(400,300);

			main_window = new MainWindow();
			script_manager = new ScriptManagerWindow();
			// job_view = new JobManager(this);

			WindowPos.setWindowsPosToCenter(main_window);
			WindowPos.setWindowsPosToCenter(script_manager);
			WindowPos.setWindowsPosToCenter(log_all);
	//		WindowPos.setWindowsPosToCenter(job_view);

			PropertyManager.load(getApp(),IniFileName);
			App.buf_list.sortByOrder();
			App.buf_list.select(root_property.getInt("buf_list_last_select",0));
			App.main_window.reloadBGInfoAll();

			// スクリプト
			for(Iterator it=root_property.getList("Scripts").iterator();it.hasNext();)
			{ script_manager.loadInfo( PropertyManager.StringValue(it.next())); }
			compiled_scripts.setup();

			// 起動時に接続する？
			for(Iterator it=getChilds();it.hasNext();){
				Object o=it.next();
				if(o instanceof CTN_Conn) ((CTN_Conn)o).onLoad();
			}
			buf_list.sub_scrollRectToVisible();

			// selectの開始
			select_thread = new SelectThread();
			select_thread.start();

			os_dependence.setLF(root_property.setDefaultString( 
				"LookAndFeel"
				,UIManager.getSystemLookAndFeelClassName()
			));

			// プロパティを読んでウィンドウ位置を再現する
			WindowPos.loadWindowPos(main_window,"WindowBounds_Main");
			WindowPos.loadWindowPos(script_manager,"WindowBounds_Scripts");
			WindowPos.loadWindowPos(log_all,"WindowBounds_AllLog");
	//		WindowPos.loadWindowPos(job_view,"WindowBounds_JobManager");

			// スプリッタ位置を再現する
			WindowPos.loadSplitterPos("ChannelAndUserSplitter",main_window.ChannelAndUserSplitter);
			WindowPos.loadSplitterPos("LRSplitter",main_window.LRSplitter);
			WindowPos.loadSplitterPos("AllLogSplitter",main_window.AllLogSplitter);
			WindowPos.loadTableColPos(main_window.user_view.table,"UserViewColmnWidth");

			if(null== root_property.get("WindowBounds_AllLog"))
			{  log_all.show(); }

			main_window.neatSplitter();
			App.main_window.setVisible(true);
			App.main_window.repaint();
			
			// タイマの開始
			timer = new javax.swing.Timer(100,new ActionListener(){
			    public void actionPerformed(ActionEvent evt){
					try{
						switch(timerState){
						case 0:
							//repaintが反映された後にLFの詳細を直す必要がある
							os_dependence.setLF(root_property.setDefaultString( 
								"LookAndFeel"
								,UIManager.getSystemLookAndFeelClassName()
							));
							System.runFinalization();

						((javax.swing.Timer)evt.getSource()).setDelay(1000);
						timerState=100;break;case 100: // timer
							App.selector.timer();
							hook_manager.callTimer();

							// PONG送信
							for(Iterator it=getChilds();it.hasNext();){
								Object o=it.next();
								if(o instanceof CTN_Conn) ((CTN_Conn)o).onTimer();
							}


						break;case 1000:
							// 終了時の待機

							System.runFinalization();
							// 接続がすべて閉じられるのを待つ
							if(!selector.closeAll() ) break;
							// 接続ツリーを消す
							removeChildAll();

							logger.fine("kill selector thread...");

						timerState=1020; /* break; */ case 1020:
							// System.exitされるまで呼ばれる
							System.runFinalization();
							select_thread.kill();
							// うまくいけばselect_threadがSystem.exitを発行する
						}
					}catch(Throwable e){
						logger.log(java.util.logging.Level.SEVERE,"timerState="+timerState,e);
					}
			    }
			});
			timer.start();
		}catch(Throwable e){
			logger.log(java.util.logging.Level.SEVERE,"init",e);
		}
	}

	public void AppExit(String reason){
		/*
			実行中にjarファイルを書き換えるとクラスローダまわりでうまく動作しなくなるのに
			jarファイルをロックしてくれないのって難儀やなぁ
		*/
		try{
			timer.stop();
		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{

			// スクリプトの状況をプロパティに保存する
			root_property.set("Scripts",script_manager.getScriptInfo());
		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{

			// スプリッタとテーブルの区切り位置
			WindowPos.saveTableColPos(main_window.user_view.table,"UserViewColmnWidth");
			WindowPos.saveSplitterPos("ChannelAndUserSplitter",main_window.ChannelAndUserSplitter);
			WindowPos.saveSplitterPos("LRSplitter",main_window.LRSplitter);
			WindowPos.saveSplitterPos("AllLogSplitter",main_window.AllLogSplitter);
		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{

			// ウィンドウ位置を保存
			WindowPos.saveWindowPos(main_window,"WindowBounds_Main");
			WindowPos.saveWindowPos(script_manager,"WindowBounds_Scripts");
			WindowPos.saveWindowPos(log_all,"WindowBounds_AllLog");
	//		WindowPos.saveWindowPos(job_view,"WindowBounds_JobManager");
		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{

			App.buf_list.saveOrder();
			root_property.set("buf_list_last_select",""+App.buf_list.getSelectedIndex() );

		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{
			// スクリプトの停止
			script_manager.exit();
		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{

			compiled_scripts.unloadAll();
		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{

			// プロパティを保存する
			PropertyManager.save(getApp(),IniFileName);
		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{

			// 接続を切る
			disconnectAll();
		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{

			System.runFinalization();

		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{
			// ウィンドウを隠す
			log_all.hide();
			main_window.hide();
			script_manager.hide();
			// job_view.hide();
		}catch(Throwable e)
		{ logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e); }
		try{

			logger.info("AppExit reason="+reason);
			timerState=1000;
			timer.setDelay(300);
			timer.restart();
		}catch(Throwable e){
			logger.log(java.util.logging.Level.SEVERE,"AppExit reason="+reason,e);
		}
	}

	///////////////////////////////////////////////////////////////////
	// implements ConnTreeNode

	public boolean isApp (){ return true;}
	public String GetCaption(){ return App.getApp().GetName(); }

	public String getPopupName(){ return "popup-app";}
	public int getListLevel(){ return 0;}
	public String getTitle(){ return App.config_AppName +" - ";}

	public boolean SendFromUserInput(String a,int type){return false;}

	public void disconnectAll(){
		Vector v =new Vector();
		for(Iterator it=child.iterator();it.hasNext();) v.add( it.next() );
		for(int i=0;i<v.size();++i){
			ConnTreeNode o = (ConnTreeNode)v.get(i);
			if(o instanceof CTN_Conn) ((CTN_Conn)o).disconnect();
		}
	}

	public java.util.List getCreateParam(){
		LinkedList l=new LinkedList();
		l.add( GetCaption() );
		return l;
	}
	public static ConnTreeNode createNode(ConnTreeNode parent,java.util.List param){
		return getApp();
	}
	public void PreRemove(){ /* nothing to do */ }

	////////////////////////////////////////////////////////////
	// プロパティと変数の対応
	public Object getPropertyExtra(){ return this; }
	public void onHashSave(){}
	public void onHashLoad(){}

	public void onPropertyChanged(String name,Object value){
		if(name.equals("LogSweepMax")){
			buffer.LogDocument.config_LogSweepMax =PropertyManager.IntValue(value);
			if(buffer.LogDocument.config_LogSweepMax<1)buffer.LogDocument.config_LogSweepMax=1;
		}
		if(name.equals("LookAndFeel")){
			os_dependence.setLF((String)value);
		}
	}

	public void setDefaultProperty(){
		property.setDefaultBoolean("ActivateBufferOnCreate",true);
		property.setDefaultBoolean("LogCRLF",true);
		property.setDefaultBoolean("RightTree",true);
		property.setDefaultBoolean("TopTree",true);
		property.setDefaultBoolean("AlloLogInMain",false);
		property.setDefaultBoolean("alllog_top",false);
		property.setDefaultBoolean("HorizontalLayout",true);
		property.setDefaultString("LogFileEncoding","UTF-8");
		property.setDefaultString("CommandSlash","/");
		property.setDefaultString("LogSweepMax","2000");

		property.setDefaultString("TextEditorPath","notepad.exe");
		property.setDefaultString("FolderViewerPath","cmd.exe /C start");
		property.setDefaultString("BrowserName","cmd.exe /C start");
	}

	//////////////////////////////////////////////////////////////////////////
	// 選択状態の取得

	// 選択中のノードとそれに関係したノードを返す
	public static ConnTreeNode getSelected(){ return buf_list==null?null:buf_list.getSelected(); }
	public static CTN_Conn getSelectedConn(){ return (buf_list==null)? null : buf_list.getConn(); }
	public static CTN_Chan getSelectedChannel(){ return (buf_list==null)? null : buf_list.getChannel(); }
	public static CTN_Priv getSelectedPriv(){ return (buf_list==null)? null : buf_list.getPrivNode(); }

	// ノードの選択
	public static void selectTreeView(ConnTreeNode node)
		{ buf_list.select(node); }

	// ノードの再描画
	public static void redrawTreeView(ConnTreeNode node)
		{ buf_list.redrawItem(node); }

	// prefferdSizeの再評価 ノードの増減や編集の際に呼ぶ必要がある
	public static void updateBufferListItem(ConnTreeNode node)
		{ buf_list.notifySizeChange(); }

	// ノードが選択された際に主ウィンドウの表示を更新する
	String lastTitle=null;
	public void updateTopic(){
		ConnTreeNode c = buf_list.getSelected();
		if(c==null) return;
		String t=c.getTitle();
		if(!t.equals(lastTitle)) App.main_window.setTitle(lastTitle=t);
		App.main_window.labelBufferInfo.setText(c.getBufferInfo());
		App.main_window.log_view.setDocument( c.getLogDocument() );
		App.main_window.user_view.setDocument(c,c.getUserList());
	}


	//////////////////////////////////////////////////////////////////////////
	// その他

	// バッファにtextを追加する
	public static void Log(String text){
		if(text==null || text.length()==0) return;
		App app= getApp();

		logger.finest(text);

		if(app==null || app.console_doc==null) return;

		for(int i=0;i<text.length();++i){
			int lf = text.indexOf('\n',i);
			if(lf==-1) lf=text.length();
			if(lf>i){
				String s=text.substring(i,lf);
				i=lf;
				// BluntIRC バッファと すべてのメッセージに出力する
				app.console_doc.addText(App.main_window,s);
				if( app.log_all!=null 
				&&  app.log_all.doc !=null
				){
					app.log_all.doc.addText(App.main_window,s);
				}
			}
		}
	}


	/////////////////////////////////////////////////////
	/*
		メニューを開くとJTextComponent からフォーカスが失われる現象の対策 
		MenuConf.java も見ること
	*/
	static Component pre_focus;
	public static void recordPreFocus(String context){
		Component c= KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner();
		if(c==null || (c instanceof JRootPane) ) return;
		pre_focus = c;
	}
	public static Object getFocusComponent(){
		Component c= KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner();
		if(c instanceof JRootPane) c = pre_focus;
		return c;
	}
}
